import asyncio
import json
import logging
import sys
from pathlib import Path
from typing import List, Tuple, Optional

import sdto
from sdto.sdto import Options, scan, fingerprints_from, RegexFingerprint

try:
    import click
except ImportError:
    print(
        "You need to reinstall this library with 'pip install sdto[cli]' to be able to use it as a CLI tool\n",
        file=sys.stderr,
    )
    raise


@click.command(help="sdto - subdomain takeover scanner")
@click.option(
    "--target",
    "-t",
    help="domain(s) to scan",
    multiple=True,
)
@click.option("--targets-list", "-l", help="domains list in a file")
@click.option(
    "--header",
    "-H",
    type=(str, str),
    help="HTTP headers",
    multiple=True,
)
@click.option(
    "--concurrency",
    default=5,
    help="max number of concurrent requests",
    type=click.IntRange(1, 100),
)
@click.option(
    "--timeout",
    "-T",
    help="HTTP requests timeout (seconds)",
    type=click.FloatRange(0, 120, min_open=True),
)
@click.option("--proxy", "-p", help="proxy to use")
@click.option("--output", "-o", help="output to")
@click.option(
    "--output-format",
    "-F",
    type=click.Choice(["txt", "json"], case_sensitive=False),
    default="txt",
    help="output format. either 'txt' or 'json'",
)
@click.option("--fingerprints", "-f", help="fingerprints file path")
@click.option(
    "--no-ssl",
    is_flag=True,
    default=True,
)
@click.option("-v", "--verbose", count=True)
def main(
    target: List[str],
    header: List[Tuple[str, str]],
    concurrency: int,
    verbose: int,
    proxy: Optional[str],
    output: Optional[str],
    output_format: str,
    targets_list: Optional[str],
    fingerprints: Optional[str],
    no_ssl: bool,
    timeout: Optional[float],
):
    log_format = "%(asctime)s - %(levelname)s - %(message)s"
    logger = logging.getLogger(__name__)
    if verbose == 0:
        logging.basicConfig(format=log_format, level=logging.WARNING)
    elif verbose == 1:
        logging.basicConfig(format=log_format, level=logging.INFO)
    else:
        logging.basicConfig(format=log_format, level=logging.DEBUG)
    options = Options(
        domains=domains_from(targets_list, target),
        headers=headers_from(header),
        concurrency=concurrency,
        verbosity=verbose,
        proxy=proxy,
        fingerprints=fingerprints_from(fingerprints),
        ssl=not no_ssl,
        timeout=timeout,
    )
    print_banner(sdto.__version__)
    matches = asyncio.run(scan(options, logger=logger))
    write(matches, output_format, output)


def print_banner(version: str):
    p = lambda t: print(t, file=sys.stderr)
    p("\n")
    p("   _______  __________ ")
    p("  / __/ _ \\/_  __/ __ \\")
    p(" _\\ \\/ // / / / / /_/ /")
    p(f"/___/____/ /_/  \\____/\t\t{version}")
    p(f"\t\t\tscanfactory.io")
    p(f"https://github.com/scanfactory/sdto")
    p("\n")


def write(
    matches: List[Tuple[str, RegexFingerprint]],
    fmt: str,
    output_to: Optional[str],
):
    if not output_to:
        print(to_string(matches, fmt))
    else:
        Path(output_to).write_text(to_string(matches, fmt))


def to_string(matches: List[Tuple[str, RegexFingerprint]], fmt: str) -> str:
    if fmt == "json":
        return json.dumps(
            [{"target": target, "match": f.name} for target, f in matches]
        )
    return "\n".join(
        f"✨✨✨ Potential takeover: [{target}] -> {f.name} ✨✨✨" for target, f in matches
    )


def headers_from(headers_param: List[Tuple[str, str]]) -> List[Tuple[str, str]]:
    headers = [(n.lower(), v) for n, v in headers_param]
    names = frozenset(h for h, _ in headers)
    if "user-agent" not in names:
        headers.append(
            (
                "user-agent",
                "text/html,application/xhtml+xml,application/xml;q=0.9,image/webp,*/*;q=0.8",
            )
        )
    if "accept-language" not in names:
        headers.append(("accept-language", "en-US,en;q=0.5"))
    return headers


def domains_from(domains_path: str, domains: List[str]) -> List[str]:
    if domains_path:
        return [i.strip() for i in Path(domains_path).read_text().split() if i.strip()]
    if domains:
        return domains
    raise TypeError("Either --domain or --domains-list must be specified")


if __name__ == "__main__":
    main()



import asyncio
import dataclasses
import json
import logging
import re
from pathlib import Path
from typing import List, Optional, Tuple, Pattern, Dict, Any
from urllib.parse import urlparse

import aiohttp

default_fingerprints = {
    "AWS/S3": {"pattern": r"The specified bucket does not exist"},
    "BitBucket": {"pattern": r"Repository not found"},
    "Github": {"pattern": r"There isn\'t a Github Pages site here\."},
    "Shopify": {"pattern": r"Sorry\, this shop is currently unavailable\."},
    "Fastly": {"pattern": r"Fastly pattern\: unknown domain\:"},
    "Ghost": {
        "pattern": r"The thing you were looking for is no longer here\, or never was"
    },
    "Heroku": {
        "pattern": r"no-such-app.html|<title>no such app</title>|herokucdn.com/pattern-pages/no-such-app.html"
    },
    "Pantheon": {
        "pattern": r"The gods are wise, but do not know of the site which you seek."
    },
    "Tumbler": {
        "pattern": r"Whatever you were looking for doesn\'t currently exist at this address."
    },
    "Wordpress": {"pattern": r"Do you want to register"},
    "TeamWork": {"pattern": r"Oops - We didn\'t find your site."},
    "Helpjuice": {"pattern": r"We could not find what you\'re looking for."},
    "Helpscout": {"pattern": r"No settings were found for this company:"},
    "Cargo": {"pattern": r"<title>404 &mdash; File not found</title>"},
    "Uservoice": {"pattern": r"This UserVoice subdomain is currently available!"},
    "Surge": {"pattern": r"project not found"},
    "Intercom": {
        "pattern": r"This page is reserved for artistic dogs\.|Uh oh\. That page doesn\'t exist</h1>"
    },
    "Webflow": {
        "pattern": r"<p class=\"description\">The page you are looking for doesn\'t exist or has been moved.</p>"
    },
    "Kajabi": {"pattern": r"<h1>The page you were looking for doesn\'t exist.</h1>"},
    "Thinkific": {
        "pattern": r"You may have mistyped the address or the page may have moved."
    },
    "Tave": {"pattern": r"<h1>pattern 404: Page Not Found</h1>"},
    "Wishpond": {"pattern": r"<h1>https://www.wishpond.com/404?campaign=true"},
    "Aftership": {
        "pattern": r"Oops.</h2><p class=\"text-muted text-tight\">The page you\'re looking for doesn\'t exist."
    },
    "Aha": {"pattern": r"There is no portal here \.\.\. sending you back to Aha!"},
    "Tictail": {
        "pattern": r"to target URL: <a href=\"https://tictail.com|Start selling on Tictail."
    },
    "Brightcove": {
        "pattern": r"<p class=\"bc-gallery-pattern-code\">pattern Code: 404</p>"
    },
    "Bigcartel": {"pattern": r"<h1>Oops! We couldn&#8217;t find that page.</h1>"},
    "ActiveCampaign": {"pattern": r"alt=\"LIGHTTPD - fly light.\""},
    "Campaignmonitor": {
        "pattern": r"Double check the URL or <a href=\"mailto:help@createsend.com"
    },
    "Acquia": {
        "pattern": r"The site you are looking for could not be found.|If you are an Acquia Cloud customer and expect to see your site at this address"
    },
    "Proposify": {
        "pattern": r"If you need immediate assistance, please contact <a href=\"mailto:support@proposify.biz"
    },
    "Simplebooklet": {
        "pattern": r"We can\'t find this <a href=\"https://simplebooklet.com"
    },
    "GetResponse": {
        "pattern": r"With GetResponse Landing Pages, lead generation has never been easier"
    },
    "Vend": {"pattern": r"Looks like you\'ve traveled too far into cyberspace."},
    "Jetbrains": {"pattern": r"is not a registered InCloud YouTrack."},
    "Smartling": {"pattern": r"Domain is not configured"},
    "Pingdom": {"pattern": r"pingdom"},
    "Tilda": {"pattern": r"Domain has been assigned"},
    "Surveygizmo": {"pattern": r"data-html-name"},
    "Mashery": {"pattern": r"Unrecognized domain <strong>"},
    "Divio": {"pattern": r"Application not responding"},
    "feedpress": {"pattern": r"The feed has not been found."},
    "readme": {"pattern": r"Project doesnt exist... yet!"},
    "statuspage": {"pattern": r"You are being <a href=\'https>"},
    "zendesk": {"pattern": r"Help Center Closed"},
    "worksites.net": {"pattern": r"Hello! Sorry, but the webs>"},
    "wix": {"pattern": r"Looks Like This Domain Isn\'t Connected To A Website Yet!"},
    "airee": {"pattern": r"Ошибка 402\. Сервис"},
}


@dataclasses.dataclass()
class RegexFingerprint:
    name: str
    pattern: Pattern
    process_200: bool


@dataclasses.dataclass
class Options:
    domains: List[str]
    headers: List[Tuple[str, str]]
    concurrency: int
    verbosity: int
    proxy: str
    fingerprints: List[RegexFingerprint]
    ssl: bool
    timeout: Optional[float]


async def token_bucket(concurrency: int) -> asyncio.Queue:
    tokens = asyncio.Queue(maxsize=-1)
    for _ in range(concurrency):
        await tokens.put(None)
    return tokens


async def scan(
    options: Options, logger: logging.Logger = logging.getLogger(__name__)
) -> List[Tuple[str, RegexFingerprint]]:
    results = []
    tasks = []
    tokens = await token_bucket(options.concurrency)
    async with aiohttp.ClientSession(
        headers=options.headers,
        timeout=aiohttp.ClientTimeout(total=options.timeout)
        if options.timeout
        else aiohttp.client.sentinel,
    ) as cs:
        for d in options.domains:
            await tokens.get()
            tasks.append(
                asyncio.create_task(
                    scan_one(
                        cs,
                        d,
                        options.ssl,
                        options.fingerprints,
                        options.proxy,
                        tokens,
                        results,
                        logger,
                    )
                )
            )
        await asyncio.wait(tasks, return_when=asyncio.ALL_COMPLETED)
    return results


def url_of(t: str, ssl: bool):
    parsed = urlparse(t)
    if parsed.scheme in ("http", "https"):
        return t
    if not parsed.scheme:
        return f"https://{t}/" if ssl else f"http://{t}/"
    raise ValueError(f"Bad url scheme in {t!r}")


async def scan_one(
    cs: aiohttp.ClientSession,
    target: str,
    ssl: bool,
    fingerprints: List[RegexFingerprint],
    proxy: Optional[str],
    tokens: asyncio.Queue,
    results: List[Tuple[str, RegexFingerprint]],
    logger: logging.Logger,
):

    logger.info(f"Checking {target!r}...")
    try:
        match = await check_target(cs, target, ssl, proxy, fingerprints)
        if match:
            results.append((target, match))
    except Exception as e:
        logger.exception(f"Got exception while checking {target!r}: {e!r}", exc_info=e)
    else:
        logger.info(
            f"Finished {target!r}. {'Matched %s' % match.name if match else 'No matches'}."
        )
    finally:
        await tokens.put(None)


def fingerprints_from(path: Optional[str]) -> List[RegexFingerprint]:
    if path:
        fingerprints = json.loads(Path(path).read_text())
    else:
        fingerprints = default_fingerprints
    try:
        return parsed_fingerprints(fingerprints)
    except Exception as e:
        raise ValueError("Bad fingerprints format") from e


async def check_target(
    cs: aiohttp.ClientSession,
    target: str,
    ssl: bool,
    proxy: Optional[str],
    fingerprints: List[RegexFingerprint],
) -> Optional[RegexFingerprint]:
    async with cs.get(url_of(target, ssl), proxy=proxy) as resp:
        try:
            text = await resp.text()
        except UnicodeDecodeError:
            return None
    return find_match(fingerprints, text, resp.status)


def find_match(
    fingerprints: List[RegexFingerprint], text: str, status_code: int
) -> Optional[RegexFingerprint]:
    for f in fingerprints:
        if (status_code in range(201, 599) or f.process_200) and f.pattern.findall(
            text
        ):
            return f
    return None


def parsed_fingerprints(f: Dict[str, Any]) -> List[RegexFingerprint]:
    return [
        RegexFingerprint(
            name,
            re.compile(data["pattern"], re.IGNORECASE),
            data.get("process_200", False),
        )
        for name, data in f.items()
    ]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['sdto']

package_data = \
{'': ['*']}

install_requires = \
['aiohttp>=3.8.3,<4.0.0']

setup_kwargs = {
    'name': 'sdto',
    'version': '0.1.3',
    'description': 'Subdomain takeover finder',
    'long_description': '[![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/psf/black)\n[![Downloads](https://pepy.tech/badge/sdto)](https://pepy.tech/project/sdto)\n\n# sdto - subdomain takeover finder\n\nSubdomain takeover scanner.\n\n[What is subdomain takeover?](https://labs.detectify.com/2014/10/21/hostile-subdomain-takeover-using-herokugithubdesk-more/)\n\n\n## Supported Services\n\n```\nacquia\nactivecampaign\naftership\naha\nairee\naws/s3\nbigcartelbitbucket\nbrightcove\ncampaignmonitor\ncargo\ncloudfront\ndesk\nfastly\nfeedpress\ngetresponse\nghost\ngithub\nhelpjuice\nhelpscout\nheroku\nintercom\njetbrains\nkajabi\nmashery\npantheon\npingdom\nproposify\nreadme\ns3bucket\nshopify\nsimplebooklet\nsmartling\nsmugmug\nstatuspage\nsurge\nsurveygizmo\ntave\nteamwork\nthinkific\ntictail\ntilda\ntumbler\nunbounce\nuservoice\nvend\nwebflow\nwishpond\nwix\nwordpress\nworksites.net\nzendesk\n```\n## Installation:\n\n\nto use as python library\n```shell\npip install sdto\n```\n\nto use as a CLI tool\n\n```shell\npip install sdto[cli]\n```\n\n\n**or:**\n```shell\ngit clone https://github.com/scanfactory/sdto.git\ncd sdto\npoetry install\n```\n## Usage as a CLI tool\n\nExamples:\n\n```shell\npython3 -m sdto -t www.domain.com\npython3 -m sdto -t www.domain.com -f path/to/custom-fingerprints-file.json\npython3 -m sdto -t https://www.domain.com/\npython3 -m sdto -t http://www.domain.com/\npython3 -m sdto -t www.domain.com --no-ssl\npython3 -m sdto -t www.domain.com -v --timeout 30\npython3 -m sdto -t www.domain.com -H "user-agent" "your-custom-user-agent" -H "another-header" "header-value"\npython3 -m sdto -t www.domain.com -F json\npython3 -m sdto -t www.domain.com -o output.txt\npython3 -m sdto -t www.domain.com -F json -o output.json\npython3 -m sdto -t www.domain.com -F txt -o output.txt\npython3 -m sdto -t www.domain.com -p http://127.0.0.1:8080 \npython3 -m sdto -l subdomains-list.txt\n```\n\n### Docker support\n\nBuild the image:\n\n```\ndocker build -t sdto .\n```\n\nRun the container:\n\n```\ndocker run -it --rm sdto -t www.domain.com -v\n```\n\n\n### Using custom fingerprints\n\nYou can specify custom fingerprints file via `-f path/to/file.json` parameter.\nThe expected json file format:\n```json\n{\n  "AWS/S3": {"pattern": "The specified bucket does not exist"},\n  "BitBucket": {"pattern": "Repository not found"},\n  "Fastly": {"pattern": "Fastly pattern\\\\: unknown domain\\\\:", "process_200": true}\n}\n```\nNote that `pattern` value is expected to be a python regexp.\n\n## Usage as a python library\n\nExample:\n\n```python\nimport re\n\nfrom aiohttp import ClientSession\nfrom sdto import check_target, RegexFingerprint\n\n\nasync def main():\n    async with ClientSession() as cs:\n        fingerprint = await check_target(\n            cs=cs,\n            target="sub.domain.com",\n            ssl=True,\n            proxy=None,\n            fingerprints=[\n                RegexFingerprint(\n                    "Github", \n                    re.compile(r"There isn\\\'t a Github Pages site here\\."),\n                    process_200=False,\n                )\n            ]\n        )\n        if not fingerprint:\n            print("No match")\n        else:\n            print(fingerprint.name)\n\n```\n',
    'author': 'godpleaseno',
    'author_email': 'zfrty@protonmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/scanfactory/sdto',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
